"""
Utils for reverse protobuf
"""
from typing import Tuple

import grpc

from cygrpc.gateway import errors
from cygrpc.protobuf_to_dict import protobuf_to_dict, dict_to_protobuf
from cygrpc.utils import extractor

registry = {}


class RegisteringType(type):
    def __init__(cls, name, bases, attrs):
        for key, val in attrs.iteritems():
            properties = getattr(val, 'route', None)
            if properties is not None:
                registry['%s.%s' % (name, key)] = properties


class ReverseStub:
    """
    Create a stub instance for rest gateway from pb2_grpc compiled module
    """

    def __init__(self, grpc_host, grpc_port):
        self.stubs = dict()
        self.server_host = grpc_host
        self.server_port = grpc_port
        self.channel = None

    def add_service(self, pb2_grpc, impl):

        _stub = extractor.extract_stub_class(pb2_grpc)
        methods = extractor.extract_servicer_methods_names(pb2_grpc)
        module = extractor.extract_module(pb2_grpc)
        requests_responses = extractor.extract_requests_responses(pb2_grpc, methods=methods)
        self.stubs[impl.__class__.__name__.strip()] = dict()
        for method in methods:
            self.stubs[impl.__class__.__name__.strip()][method] = {
                "_stub": _stub,
                "stub": None,
                "module": module,
                "requests_response": requests_responses[method.strip()]
            }

    def init_channel(self):
        _host = "localhost" if self.server_host == "0.0.0.0" else self.server_host
        self.channel = grpc.insecure_channel(f"{_host}:{self.server_port}")
        current = ""
        temp_stub = None
        for impl in self.stubs:
            for method in self.stubs[impl]:
                if impl != current:
                    temp_stub = self.stubs[impl][method]["_stub"](self.channel)
                # self.stubs[impl][method]["stub"] = temp_stub
                del self.stubs[impl][method]["_stub"]
                self.stubs[impl][method]["stub_method"] = getattr(temp_stub, method)

    def execute(self, service: str, method: str, payload: dict, metadata: tuple = ()) -> Tuple[dict, dict]:
        """
        :param service:
        :param method:
        :param payload:
        :param metadata:
        :return: response body , metadata headers
        """
        request = None
        try:
            request = dict_to_protobuf(
                self.stubs[service.strip()][method.strip()]["requests_response"]["request"](),
                payload
            )
        except Exception as err:
            raise errors.HttpErrBadRequest(str(err))
        response, call = self.stubs[service.strip()][method.strip()]["stub_method"].with_call(request,
                                                                                              metadata=metadata)
        metadata_response = {}
        for key, value in call.trailing_metadata():
            metadata_response[key] = value
        return protobuf_to_dict(response, use_enum_labels=True, including_default_value_fields=True), metadata_response



from abc import ABC, abstractmethod

import grpc


def _unary_unary_rpc_terminator(code, details):
    def terminate(ignored_request, context):
        context.abort(code, details)

    return grpc.unary_unary_rpc_method_handler(terminate)


class CyGrpcInterceptor(ABC, grpc.ServerInterceptor):

    @classmethod
    def get_metadata_value(cls, context, key_name: str):
        for item in context.invocation_metadata:
            if item.key == key_name:
                return item.value
        return None

    @classmethod
    def get_service_name(cls, handler_call_details):
        return handler_call_details[0].split("/")[1]

    @classmethod
    def get_method_name(cls, handler_call_details):
        return handler_call_details[0].split("/")[-1]

    @abstractmethod
    def intercept_service(self, continuation, handler_call_details) -> grpc.RpcMethodHandler:
        pass

    @classmethod
    def terminator(cls, code, details):
        return _unary_unary_rpc_terminator(code, details)

    @classmethod
    def on_success(cls, continuation, handler_call_details) -> grpc.RpcMethodHandler:
        return continuation(handler_call_details)

    @classmethod
    def on_failed(cls, status_code: grpc.StatusCode, details: str) -> grpc.RpcMethodHandler:
        return cls.terminator(grpc.StatusCode.UNAUTHENTICATED, "Validate authentication failed.")



"""
Extract data from _pb2_grpc compiled file

"""
import importlib as _importlib
import inspect as _inspect
import typing as _typing


def extract_stub_class(pb2_grpc) -> _typing.ClassVar:
    """
    Extract the stub class from pb2_grpc compiled
    and return class for instance

    :param pb2_grpc:
    :return:
    """
    members = dir(pb2_grpc)
    for name in members:
        if name.find("Stub") > 0:
            _stub = getattr(pb2_grpc, name)
            return _stub
    return None


def extract_add_server(pb2_grpc):
    """
    Extract add to server function from pb2_grpc module
    :param pb2_grpc:
    :return: function for add_to_server
    :type: function
    """
    members = dir(pb2_grpc)
    for name in members:
        if name.startswith("add_") and name.endswith("_to_server"):
            _add_to_server = getattr(pb2_grpc, name)
            return _add_to_server
    return None


def extract_module(pb2_grpc):
    """
    Extract module pb2 that contains all response and request objects used
    in grpc service methods.

    is equals to: import example_pb2
    :param pb2_grpc:
    :return:
    """
    lines = _inspect.getsource(pb2_grpc).split("\n")
    for line in lines:
        if line.count("from") > 0 and line.count("as") >= 0 and line.count("_pb2") > 0:
            full_module_name = line.split(" as ")[0]
            full_module_name = full_module_name.replace(" import ", ".")
            full_module_name = full_module_name.replace("from ", "")
            return _importlib.import_module(full_module_name)


def extract_stub_methods_names(grpc_stub_class) -> _typing.List[str]:
    """
    Extract the stub methods from pb2_grpc compiled
    :param pb2_grpc:
    :return:
    """
    return list(grpc_stub_class.__dict__.keys())


def extract_servicer_methods_names(pb2_grpc):
    members = dir(pb2_grpc)
    for name in members:
        if name.endswith("Servicer"):
            servicer = getattr(pb2_grpc, name)
            return [x for x in servicer.__dict__.keys() if not str(x).startswith("__")]
    return None


def extract_requests_responses(pb2_grpc, **kwargs) -> _typing.Dict:
    """
    Extract all request and response from pb2_grpc and return a dictionary
    with the next extructure:
    {
        "method":{
            "grpc_path": path,
            "request": obj_request,
            "response": obj_response
        },
        ...
    }
    :param pb2_grpc: pb2_grpc modules
    :keyword methods_list: names for methods list. if not set auto extract from pb2_grpc
    :type: list
    :keyword module: _pb2 module, if not set auto extract from pb2_grpc
    :type: module
    :return:
    """
    lines = _inspect.getsource(pb2_grpc).split("\n")
    # print("lines", len(lines))
    methods = kwargs["methods"] if "methods" in kwargs else \
        extract_servicer_methods_names(pb2_grpc)
    module = kwargs["module"] if "module" in kwargs else \
        extract_module(pb2_grpc)
    requests_responses = dict()
    processed = list()
    for numb, line in enumerate(lines):
        for method in methods:
            if line.count(f"{method}") > 0 and method not in processed:
                processed.append(method)
                path = __sanitize_path(lines[numb + 1])
                request = __extract_protobuf_request_object_from_str(lines[numb + 2])
                response = __extract_protobuf_response_object_from_str(lines[numb + 3])
                obj_request = getattr(module, request)
                obj_response = getattr(module, response)
                requests_responses[method] = {
                    "grpc_path": path,
                    "request": obj_request,
                    "response": obj_response
                }
    return requests_responses


def __sanitize_path(line: str):
    return line.replace("'", "").replace(",", "").replace('"', "")


def __extract_protobuf_request_object_from_str(line: str):
    line = line.strip().replace(",", "").replace("request_serializer=", "").replace(".SerializeToString", "")
    line = line.split(".")
    return line[-1]


def __extract_protobuf_response_object_from_str(line: str):
    line = line.strip().replace(",", "").replace("response_deserializer=", "").replace(".FromString", "")
    # print(line)
    line = line.split(".")
    return line[-1]



import json


class HttpErr(Exception):
    status: int = 500
    value: str = "Internal Server Error"
    details = ""

    def __init__(self, details):
        self.details = details

    def __str__(self):
        return json.dumps(self.dict())

    def dict(self):
        return {"status": self.status, "message": self.value, "details": self.details}


# ===  4XX - Client errors ===


class HttpErrBadRequest(HttpErr):
    status = 400
    value = "Bad request"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrUnauthorized(HttpErr):
    status = 401
    value = "Unauthorized"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrPaymentRequired(HttpErr):
    status = 402
    value = "Payment Required"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrForbidden(HttpErr):
    status = 403
    value = "Forbidden"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrNotFound(HttpErr):
    status = 404
    value = "Not Found"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrMethodNotAllowed(HttpErr):
    status = 405
    value = "Method Not Allowed"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrNotAcceptable(HttpErr):
    status = 406
    value = "Not Acceptable"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrProxyAuthenticationRequired(HttpErr):
    status = 407
    value = "Proxy Authentication Required"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrRequestTimeout(HttpErr):
    status = 408
    value = "Request Timeout"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrConflict(HttpErr):
    status = 409
    value = "Conflict"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrGone(HttpErr):
    status = 410
    value = "Gone"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrLengthRequired(HttpErr):
    status = 411
    value = "Length Required"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrPreconditionFailed(HttpErr):
    status = 412
    value = "Precondition Failed"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrPayloadTooLarge(HttpErr):
    status = 413
    value = "Payload Too Large"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrURITooLong(HttpErr):
    status = 414
    value = "URI Too Long"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrUnsupportedMediaType(HttpErr):
    status = 415
    value = "Unsupported Media Type"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrRangeNotSatisfiable(HttpErr):
    status = 416
    value = "Range Not Satisfiable"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrExpectationFailed(HttpErr):
    status = 417
    value = "Expectation Failed"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrUnprocessableEntity(HttpErr):
    status = 422
    value = "Unprocessable Entity"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrLocked(HttpErr):
    status = 423
    value = "Locked"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrFailedDependency(HttpErr):
    status = 424
    value = "Locked"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrTooEarly(HttpErr):
    status = 425
    value = "Too Early"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrUpgradeRequired(HttpErr):
    status = 426
    value = "Upgrade Required"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrPreconditionRequired(HttpErr):
    status = 428
    value = "Precondition Required"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrTooManyRequests(HttpErr):
    status = 428
    value = "Too Many Requests"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrRequestHeaderFieldsTooLarge(HttpErr):
    status = 431
    value = "Request Header Fields Too Large"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrUnavailableForLegalReasons(HttpErr):
    status = 431
    value = "Unavailable For Legal Reasons"

    def __init__(self, details: str = ""):
        super().__init__(details)


# ===  5XX - Server errors ===

class HttpErrInternalServerError(HttpErr):
    status = 500
    value = "Internal Server Error"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrNotImplemented(HttpErr):
    status = 501
    value = "Not Implemented"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrBadGateway(HttpErr):
    status = 502
    value = "Bad Gateway"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrServiceUnavailable(HttpErr):
    status = 503
    value = "Service Unavailable"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrGatewayTimeout(HttpErr):
    status = 504
    value = "Gateway Timeout"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrHTTPVersionNotSupported(HttpErr):
    status = 505
    value = "HTTP Version Not Supported"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrVariantAlsoNegotiates(HttpErr):
    status = 506
    value = "Variant Also Negotiates "

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrInsufficientStorage(HttpErr):
    status = 507
    value = "Insufficient Storage"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrLoopDetected(HttpErr):
    status = 508
    value = "Loop Detected"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrNotExtended(HttpErr):
    status = 510
    value = "Not Extended"

    def __init__(self, details: str = ""):
        super().__init__(details)


class HttpErrNetworkAuthenticationRequired(HttpErr):
    status = 511
    value = "Network Authentication Required"

    def __init__(self, details: str = ""):
        super().__init__(details)


def http_status_from_grpc_code(status_code, details="") -> HttpErr:
    import grpc
    if grpc.StatusCode.UNKNOWN == status_code:
        return HttpErrInternalServerError(details)
    if grpc.StatusCode.CANCELLED == status_code:
        return HttpErrRequestTimeout(details)
    elif grpc.StatusCode.INVALID_ARGUMENT == status_code:
        return HttpErrBadRequest(details)
    elif grpc.StatusCode.DEADLINE_EXCEEDED == status_code:
        return HttpErrGatewayTimeout(details)
    elif grpc.StatusCode.NOT_FOUND == status_code:
        return HttpErrNotFound(details)
    elif grpc.StatusCode.ALREADY_EXISTS == status_code:
        return HttpErrConflict(details)
    elif grpc.StatusCode.PERMISSION_DENIED == status_code:
        return HttpErrForbidden(details)
    elif grpc.StatusCode.UNAUTHENTICATED == status_code:
        return HttpErrUnauthorized(details)
    elif grpc.StatusCode.RESOURCE_EXHAUSTED == status_code:
        return HttpErrTooManyRequests(details)
    elif grpc.StatusCode.FAILED_PRECONDITION == status_code:
        # Note, this deliberately doesn't translate to the similarly named 412 Precondition Failed HTTP response status
        return HttpErrBadRequest(details)
    elif grpc.StatusCode.ABORTED == status_code:
        return HttpErrConflict(details)
    elif grpc.StatusCode.OUT_OF_RANGE == status_code:
        return HttpErrBadRequest(details)
    elif grpc.StatusCode.UNIMPLEMENTED == status_code:
        return HttpErrNotImplemented(details)
    elif grpc.StatusCode.UNAVAILABLE == status_code:
        return HttpErrServiceUnavailable(details)
    elif grpc.StatusCode.DATA_LOSS == status_code:
        return HttpErrInternalServerError(details)
    else:
        return HttpErrInternalServerError(details)



from cygrpc.gateway.middleware.base import BasePreMiddleware


class CygrpcAuthMiddleware(BasePreMiddleware):

    def __init__(self, function):
        self.function = function

    def process(self, route: dict, request: dict, header: dict) -> None:
        self.function(route, request, header)



"""
Http router

"""
import json
import logging as _logging
import os
import traceback

import bottle
import grpc
from bottle import request, response
from truckpad.bottle.cors import CorsPlugin, enable_cors

from cygrpc.gateway import errors
from cygrpc.gateway.middleware import CygrpcLogMiddleware
from cygrpc.gateway.proxy.middleware import MiddlewareManager
from cygrpc.gateway.proxy.reverse_stub import ReverseStub

debuger = False
# increase payload size
bottle.BaseRequest.MEMFILE_MAX = 1024 * 1024 * 100

_logging.basicConfig(
    level=os.getenv("LOGLEVEL", "INFO"),
    format='%(levelname)s | %(asctime)s - %(name)s - %(message)s')


class Router:
    __SINGLE_INSTANCE = None

    def __new__(cls, *args, **kwargs):
        if not Router.__SINGLE_INSTANCE:
            _router = Router.__Router()
            cls.__SINGLE_INSTANCE = _router
        return cls.__SINGLE_INSTANCE

    class __Router:
        reverse_stub: ReverseStub = None

        def add_route(self, service: str, grpc_method: str, url_path: str, http_method: str):
            # TODO:: add cors to optional
            @enable_cors
            @bottle.route(url_path, method=['OPTIONS', http_method])
            def intercent(**args):
                # print(bottle.request.route)
                # get request json payload
                request_json = bottle.request.json if request.json is not None else {}
                # routing dict
                route = {"method": bottle.request.route.method, "rule": bottle.request.route.rule, "path": bottle.request.path}
                # merge json payload with url params
                payload = {**args, **request_json}
                # extract request headers
                request_headers = {}
                for key in dict(request.headers):
                    request_headers[key.lower()] = dict(request.headers)[key]
                # dictionary with all request content
                request_obj = {"route": bottle.request.path, "method": http_method, "payload": payload,
                               "headers": request_headers}
                # request headers base
                response_headers = {"Content-Type": "application/json", "status": 200}
                try:
                    # execute all pre middleware http
                    for middleware in MiddlewareManager().global_pre_middleware:
                        middleware.process(route, payload, request_headers)
                    # parse request headers to metadata
                    head_to_metadata = [(str(key).lower(), request_headers[key]) for key in request_headers]
                    head_to_metadata.append(("interface", "HTTP"))
                    head_to_metadata = tuple(head_to_metadata)
                    # call stub ...
                    result, result_headers = Router().reverse_stub.execute(service, grpc_method, payload=payload,
                                                                           metadata=head_to_metadata)
                    # merge result metadata headers in response headers base
                    for key in result_headers:
                        response_headers[key] = result_headers[key]
                    # keep status code
                    status = response_headers["status"]
                    # execute all pos middleware
                    for middleware in MiddlewareManager().global_pos_middleware:
                        middleware.on_success(request_obj, result, response_headers)
                    # put all headers in response object
                    for key in response_headers:
                        response.set_header(key, response_headers[key])
                    # set  status from headers if is present
                    if "status" in response_headers:
                        status = int(response_headers["status"])
                    return bottle.HTTPResponse(status=status, body=json.dumps(result), headers=response_headers)
                except RuntimeError as err:
                    http_error = errors.HttpErrBadRequest(str(err.__cause__ if err.__cause__ is not None else err))
                    response_headers['status'] = http_error.status
                    error_response = http_error.dict()
                    for middleware in MiddlewareManager().global_pos_middleware:
                        middleware.on_error(request_obj, error_response, response_headers, err)
                    status = response_headers['status'] if "status" in response_headers else http_error.status
                    return bottle.HTTPResponse(status=status, body=json.dumps(error_response), headers=response_headers)
                except grpc.RpcError as err:
                    http_error = errors.http_status_from_grpc_code(err.code(), err.details())
                    response_headers['status'] = http_error.status
                    error_response = http_error.dict()
                    for middleware in MiddlewareManager().global_pos_middleware:
                        middleware.on_error(request_obj, error_response, response_headers, err)
                    status = response_headers['status'] if "status" in response_headers else http_error.status
                    return bottle.HTTPResponse(status=status, body=json.dumps(error_response), headers=response_headers)
                except errors.HttpErr as http_error:
                    response_headers['status'] = http_error.status
                    error_response = http_error.dict()
                    for middleware in MiddlewareManager().global_pos_middleware:
                        middleware.on_error(request_obj, error_response, response_headers, http_error)
                    status = response_headers['status'] if "status" in response_headers else http_error.status
                    return bottle.HTTPResponse(status=status, body=json.dumps(error_response), headers=response_headers)
                except Exception as err:
                    error_response = {}
                    traceback.print_exc()
                    status = response_headers['status'] = 500
                    for middleware in MiddlewareManager().global_pos_middleware:
                        middleware.on_error(request_obj, error_response, response_headers, err)
                    if "status" in response_headers:
                        status = int(response_headers["status"])
                    return bottle.HTTPResponse(status=status, body=json.dumps(error_response), headers=response_headers)

        def run_serve(self, port=3000, debug=False):
            debuger = debug
            MiddlewareManager().add_pos_middleware(CygrpcLogMiddleware)
            _logging.info(f"starting http serve on port: {port}")
            bottle.install(CorsPlugin(origins=['*']))
            bottle.run(host='0.0.0.0', port=port, debug=debug, quiet=True, server='paste')
            exit(0)



import inspect
from typing import List

from cygrpc.gateway.middleware import base


class MiddlewareManager:
    __SINGLE_INSTANCE = None

    def __new__(cls, *args, **kwargs):
        if cls.__SINGLE_INSTANCE is not None:
            return cls.__SINGLE_INSTANCE
        else:
            cls.__SINGLE_INSTANCE = cls.__MiddlewareManager()
            return cls.__SINGLE_INSTANCE

    class __MiddlewareManager:

        def __init__(self):
            self.global_pre_middleware: List[base.BasePreMiddleware] = list()
            self.global_pos_middleware: List[base.BasePosMiddleware] = list()
            self.routes_middleware = {}

        def add_pre_middleware(self, middlewares):
            middleware_list = middlewares
            if not isinstance(middlewares, list):
                middleware_list = [middlewares]
            for middleware_class in middleware_list:
                middleware = middleware_class() if inspect.isclass(middleware_class) else middleware_class
                if isinstance(middleware, base.BasePreMiddleware):
                    self.global_pre_middleware.append(middleware)
                else:
                    raise NotImplementedError(
                        "middleware not implement cygrc.gateway.middleware.base.BasePreMiddleware abstract class")

        def add_pos_middleware(self, middlewares):
            middleware_list = middlewares
            if not isinstance(middlewares, list):
                middleware_list = [middlewares]
            for middleware_class in middleware_list:
                middleware = middleware_class() if inspect.isclass(middleware_class) else middleware_class
                if isinstance(middleware, base.BasePosMiddleware):
                    self.global_pos_middleware.append(middleware)
                else:
                    raise NotImplementedError(
                        "middleware not implement cygrc.gateway.middleware.base.BasePosMiddleware abstract class"
                    )

        def add_route_pre_middleware(self, middleware_class, route: str):
            middleware = middleware_class() if inspect.isclass(middleware_class) else middleware_class
            if isinstance(middleware, base.BasePreMiddleware):
                if route in self.routes_middleware:
                    self.routes_middleware[route]["pre_middleware"].append(middleware)
                else:
                    self.routes_middleware[route] = {"pre_middleware": [], "pos_middleware": []}
                    self.routes_middleware[route]["pre_middleware"].append(middleware)
            else:
                raise NotImplementedError(
                    "middleware not implement cygrc.gateway.middleware.base.BasePreMiddleware abstract class"
                )

        def add_route_pos_middleware(self, middleware_class, route: str):
            middleware = middleware_class() if inspect.isclass(middleware_class) else middleware_class
            if isinstance(middleware, base.BasePosMiddleware):
                if route in self.routes_middleware:
                    self.routes_middleware[route]["pos_middleware"].append(middleware)
                else:
                    self.routes_middleware[route] = {"pre_middleware": [], "pos_middleware": []}
                    self.routes_middleware[route]["pos_middleware"].append(middleware)
            else:
                raise NotImplementedError(
                    "middleware not implement cygrc.gateway.middleware.base.BasePosMiddleware abstract class"
                )



from cygrpc.gateway.middleware.base import BasePreMiddleware, BasePosMiddleware
from cygrpc.gateway.middleware.log_middleware import CygrpcLogMiddleware
from cygrpc.gateway.middleware.pre_middlewares import CygrpcAuthMiddleware



from cygrpc.gateway.middleware.base import BasePosMiddleware


class MongoErrorsMiddleware(BasePosMiddleware):
    def on_success(self, request: dict, response: dict, header: dict) -> None:
        pass

    def on_error(self, request: dict, response: dict, header: dict, error: dict) -> None:
        pass



"""
Gateway instance
"""
import os
import threading

from cygrpc.gateway.proxy.reverse_stub import ReverseStub
from cygrpc.gateway.proxy.router import Router

PORT = 8080
ENABLE_DEBUG = True if os.getenv("ENVIRONMENT", "production").lower() == "develop" else False


def _run_http(stub_handler: ReverseStub, port: int, http_debug: bool):
    """
    stub_handler.init_channel()
    _HttpHandler.proto_handler = stub_handler
    with socketserver.TCPServer(("", PORT), _HttpHandler) as httpd:
        print("serving http at port", PORT)
        httpd.serve_forever()
    """
    stub_handler.init_channel()
    router = Router()
    router.reverse_stub = stub_handler
    router.run_serve(port=port, debug=http_debug)


class HttpGateway:
    """
    Gateway implementation
    """

    def __init__(self, grpc_host="0.0.0.0", grpc_port=50051):
        self.stub_handler = ReverseStub(grpc_host=grpc_host, grpc_port=grpc_port)

    def add_service(self, pb2_grpc, impl):
        self.stub_handler.add_service(pb2_grpc, impl)

    def start(self, port=3000, http_debug=False):

        thread = threading.Thread(target=_run_http, args=(self.stub_handler, port, http_debug,))
        thread.run()



import abc


class BasePreMiddleware(metaclass=abc.ABCMeta):
    """
    Execute action method
    """

    @abc.abstractmethod
    def process(self, route: dict, request: dict, header: dict) -> None:
        raise NotImplementedError("BasePreMiddleware.action")


class BasePosMiddleware(metaclass=abc.ABCMeta):
    """
    Pos middleware if process is done call on_success, if any except is returned call on_error
    """

    @abc.abstractmethod
    def on_success(self, request: dict, response: dict, header: dict) -> None:
        raise NotImplementedError("BasePreMiddleware.action")

    @abc.abstractmethod
    def on_error(self, request: dict, response: dict, header: dict, error: dict) -> None:
        raise NotImplementedError("BasePreMiddleware.action")



import logging
import os

from cygrpc.gateway.middleware.base import BasePosMiddleware

logging.basicConfig(
    level=os.getenv("LOGLEVEL", "INFO"),
    format='%(levelname)s | %(asctime)s - %(message)s')


class CygrpcLogMiddleware(BasePosMiddleware):
    def on_success(self, request: dict, response: dict, header: dict) -> None:
        logging.info(
            f"{request['method']} | {request['route']} | length: f{request['headers']['content-length']} | status: {header['status']}")

    def on_error(self, request: dict, response: dict, header: dict, error: Exception) -> None:
        logging.error(
            f"{request['method']} | {request['route']}  | status: {header['status']} | type: {type(error)} | error :{str(error.__cause__ if error.__cause__ is not None else error)}")



from abc import abstractmethod

import grpc

from cygrpc.middleware import CyGrpcInterceptor


class CyGrpcAuthInterceptor(CyGrpcInterceptor):
    @abstractmethod
    def auth_process(self, continuation, handler_call_details) -> bool:
        pass

    def intercept_service(self, continuation, handler_call_details):
        print("service: ", self.get_service_name(handler_call_details))
        print("method: ", self.get_method_name(handler_call_details))
        result = self.auth_process(continuation, handler_call_details)
        if result:
            return self.on_success(continuation, handler_call_details)
        else:
            return self.on_failed(grpc.StatusCode.UNAUTHENTICATED, "Validate authentication failed.")



import pathlib, os
from setuptools import setup, find_packages
# from pathlib import Path
from typing import List


def parse_requirements(filename: str) -> List[str]:
    """Return requirements from requirements file."""
    # Ref: https://stackoverflow.com/a/42033122/
    requirements: list = ""
    with open(os.getcwd() + '/' + filename) as file:
        requirements = file.readlines()
    requirements = [r.strip() for r in requirements]
    requirements = [r for r in sorted(requirements) if r and not r.startswith('#')]
    return requirements


# CyGRPC version
VERSION: str = '1.0.3'
# The directory containing this file
HERE = pathlib.Path(__file__).parent
# The text of the README file
README: str = (HERE / "README.md").read_text()

packages = find_packages()
packages.append("cygrpc")
# Setup
setup(
    name='cygrpc',
    version=VERSION,
    packages=packages,
    # package_dir={"": "src"},
    include_package_data=True,
    python_requires=">=3.6",
    url='https://github.com/cuemby/python-cygrpc',
    license='MIT',
    author='Fabio Moreno',
    author_email='fabio.moreno@cuemby.com',
    description='gRPC Micro framework',
    long_description=README,
    long_description_content_type='text/markdown',
    setup_requires=['wheel', 'twine'],
    install_requires=[
        'wheel',
        'twine',
        'bottle',
        'Paste',
        'grpcio>=1.23.0',
        'grpcio-tools>=1.23.0',
        'bottle-cors'
    ],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    entry_points={
        'console_scripts': [
            'cygrpc = cygrpc.cli:main',
        ]
    },
    #    scripts=['bin/cygrpc.py']
)



from typing import Any


def rest(url: str, method="GET", pre_request=None, pos_request=None) -> Any:
    """
    Decorator for add route to http rest interface
    :param pos_request: array of Middleware's that implement BasePosMiddleware class
    :param pre_request: array of Middleware's that implement BasePreMiddleware class
    :param url: Route url. example. /api/v1/service/method
    :param method: http method. GET, POST, DELETE, PUT ...
    :return: None
    """
    from cygrpc.gateway.proxy.router import Router as _Router
    from cygrpc.gateway.proxy.middleware import MiddlewareManager as _MiddlewareManager

    def inner(func):
        """
        Inject route definition to Router
        :param func:
        :return:
        """

        service = func.__qualname__.split(".")[0]
        _Router().add_route(
            service=service,
            grpc_method=func.__name__,
            url_path=url,
            http_method=method
        )
        if pre_request is not None and len(pre_request) > 0:
            _MiddlewareManager().add_route_pre_middleware(pre_request, url)
        if pos_request is not None and len(pos_request) > 0:
            _MiddlewareManager().add_route_pre_middleware(pos_request, url)
        return func

    return inner



import inspect as _inspect
import logging as _logging
import os
import time
from concurrent import futures

import grpc

from cygrpc.gateway.proxy.middleware import MiddlewareManager
from cygrpc.gateway.rest_gateway import HttpGateway
from cygrpc.utils import extractor as _extractor

_logging.basicConfig(
    level=os.getenv("LOGLEVEL", "INFO"),
    format='%(levelname)s | %(asctime)s - %(name)s - %(message)s')


class Server:
    """
    Create and run grpc server.
    """
    __SINGLE_INSTANCE = None

    def __new__(cls, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :keyword port
        :keyword host
        :keyword max_threads
        :keyword secure
        :keyword interceptors
        :return:
        """
        if not Server.__SINGLE_INSTANCE:
            Server.__SINGLE_INSTANCE = Server.__Server()
            Server.__SINGLE_INSTANCE._port = 50051 if "port" not in kwargs else kwargs["port"]
            Server.__SINGLE_INSTANCE._host = "0.0.0.0" if "host" not in kwargs else kwargs["host"]
            Server.__SINGLE_INSTANCE._max_threads = 10 if "max_threads" not in kwargs else kwargs["max_threads"]
            Server.__SINGLE_INSTANCE._secure = False if "secure" not in kwargs else kwargs["secure"]
            Server.__SINGLE_INSTANCE._interceptors = None if "interceptors" not in kwargs else kwargs["interceptors"]
            Server.__SINGLE_INSTANCE._rest_port = 3000 if "http_port" not in kwargs else kwargs["http_port"]
            Server.__SINGLE_INSTANCE._debug = False if "http_debug" not in kwargs else kwargs["http_debug"]
            Server.__SINGLE_INSTANCE.init()
        return Server.__SINGLE_INSTANCE

    class __Server:
        _grpc_server: grpc.Server = None
        _max_threads: int = 10
        _host: str = "0.0.0.0"
        _port: int = 50051
        _secure: bool = False
        _interceptors: tuple = ()
        _pb2_grpc = None
        _gateway = None
        _rest_port: int = 3000
        _debug: bool = False

        def get_grpc_server(self):
            return self._grpc_server

        def init(self):
            self._grpc_server = grpc.server(futures.ThreadPoolExecutor(max_workers=self._max_threads),
                                            interceptors=self._interceptors)
            if not self._secure:
                self._grpc_server.add_insecure_port(f"{self._host}:{self._port}")
            return self

        def add_service(self, pb2_grpc, service, rest=False):
            """
            Add service to grpc server
            :param pb2_grpc: pb2_grpc module
            :param service: service implementation
            :param rest: enable http rest gateway for service. default False
            :return:
            """

            add_to_server = _extractor.extract_add_server(pb2_grpc)
            if _inspect.isclass(service):
                service = service()
            add_to_server(service, self._grpc_server)
            if rest:
                if self._gateway is None:
                    self._gateway: HttpGateway = HttpGateway(grpc_host=self._host, grpc_port=self._port)
                self._gateway.add_service(pb2_grpc, service)

        @staticmethod
        def add_http_pre_middleware(middleware_class):
            MiddlewareManager().add_pre_middleware(middleware_class)

        @staticmethod
        def add_http_pos_middleware(middleware_class):
            MiddlewareManager().add_pos_middleware(middleware_class)

        def start(self):
            """
            Start server
            :return:
            """
            _logging.info(f"Server start in port:  {self._port}")
            self._grpc_server.start()
            if self._gateway is not None:
                time.sleep(0.15)
                self._gateway.start(port=self._rest_port, http_debug=self._debug)
            try:
                while True and self._grpc_server is not None:
                    time.sleep(3000)
            except KeyboardInterrupt as ex:
                _logging.error("Server stop.... reason: KeyboardInterrupt")
                exit(0)

        def stop(self):
            self._grpc_server.stop(0)
            exit(0)



# flake8: noqa
from cygrpc.protobuf_to_dict.convertor import (protobuf_to_dict, TYPE_CALLABLE_MAP, dict_to_protobuf,
                                               REVERSE_TYPE_CALLABLE_MAP, datetime_to_timestamp,
                                               timestamp_to_datetime, get_field_names_and_options, FieldsMissing,
                                               validate_dict_for_required_pb_fields)
